/*
  capstats - client.c
  Copyright (C) 2009 Bert Vermeulen

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <poll.h>
#include <signal.h>

#include "capstats.h"

extern int debug;
extern struct pollfd allfd[];

int server_died;
char pingsockname[128];


void ping_timeout_reached(int signum)
{

	if(debug)
		printf("ping timeout reached\n");

	unlink(pingsockname);
	alarm(0);

}


void ping_server(int signum)
{
	struct sockaddr_un pingsa, srvsa;
	socklen_t srvsa_len;
	pid_t pid;
	int sock;
	char buf[256];

	/* assume it's dead unless everything below works out */
	server_died = 1;

	/* this function should not take longer than KEEPALIVE_TIMEOUT to run. */
	signal(SIGALRM, ping_timeout_reached);
	alarm(KEEPALIVE_TIMEOUT);

	strcpy(pingsockname, PINGSOCKET);
	pid = getpid();
	snprintf(pingsockname + strlen(PINGSOCKET), sizeof(pingsockname)-1, "%d", pid);
	if(debug)
		printf("ping socket is %s\n", pingsockname);

	if( (sock = socket(AF_UNIX, SOCK_DGRAM, 0)) == -1)
	{
		if(debug)
			printf("client failed to created ping socket: %s\n", strerror(errno));
		return;
	}
	if(debug)
		printf("client created ping socket\n");

	memset(&pingsa, 0, sizeof(pingsa));
	pingsa.sun_family = AF_UNIX;
	strcpy(pingsa.sun_path, pingsockname);
	if(bind(sock, (const struct sockaddr *) &pingsa, SUN_LEN(&pingsa)) > -1)
	{
		if(debug)
			printf("client ping socket bound\n");

		srvsa.sun_family = AF_UNIX;
		strcpy(srvsa.sun_path, CMDSOCKET);
		srvsa_len = sizeof(srvsa.sun_family) + strlen(CMDSOCKET);
		if(debug)
			printf("sending ping\n");
		if(sendto(sock, "ping\n", 5, 0, (struct sockaddr *) &srvsa, SUN_LEN(&srvsa)) > -1)
		{
			if(debug)
				printf("ping sent, waiting for pong\n");
			if(recvfrom(sock, buf, sizeof(buf)-1, 0, NULL, NULL) > -1)
			{
				if(!strncasecmp(buf, "pong\n", 5))
				{
					/* everything worked out */
					server_died = 0;
					if(debug)
						printf("got pong\n");
				}
			}
		}
	}

	close(sock);
	unlink(pingsockname);

	if(server_died)
	{
		printf("server went away\n");
		alarm(0);
	}
	else
	{
		/* set up the next ping run */
		signal(SIGALRM, ping_server);
		alarm(KEEPALIVE_INTERVAL);
	}

}


int send_command(char *command)
{
	struct sockaddr_un clntsa, srvsa;
	socklen_t srvsa_len;
	pid_t pid;
	int sock, rlen, done, ret;
	char sockname[strlen(CLIENTSOCKET) + 10], buf[256];

	ret = 0;
	server_died = 0;

	strcpy(sockname, CLIENTSOCKET);
	pid = getpid();
	sprintf(sockname + strlen(CLIENTSOCKET), "%d", pid);
	if(debug)
		printf("client socket is %s\n", sockname);

	if( (sock = socket(AF_UNIX, SOCK_DGRAM, 0)) == -1)
	{
		printf("client failed to create socket: %s\n", strerror(errno));
		return(1);
	}
	if(debug)
		printf("client created command socket\n");

	memset(&clntsa, 0, sizeof(clntsa));
	clntsa.sun_family = AF_UNIX;
	strcpy(clntsa.sun_path, sockname);
	if(bind(sock, (const struct sockaddr *) &clntsa, sizeof(clntsa)) == -1)
	{
		printf("client failed to bind command socket: %s\n", strerror(errno));
		close(sock);
		return(1);
	}
	if(debug)
		printf("client command socket bound\n");

	if(debug)
		printf("sending [%s]\n", command);
	srvsa.sun_family = AF_UNIX;
	strcpy(srvsa.sun_path, CMDSOCKET);
	srvsa_len = sizeof(srvsa.sun_family) + strlen(CMDSOCKET);
	if( (sendto(sock, command, strlen(command), 0, (struct sockaddr *) &srvsa, sizeof(srvsa))) == -1)
	{
		printf("client failed to send command: %s\n", strerror(errno));
		close(sock);
		return(1);
	}

	/* set up the keepalive ping */
	siginterrupt(SIGALRM, 1);
	signal(SIGALRM, ping_server);
	alarm(KEEPALIVE_INTERVAL);

	done = 0;
	while(!done)
	{
		if( (rlen = recvfrom(sock, buf, sizeof(buf)-1, 0, NULL, 0)) == -1)
		{
			if(errno == EINTR && server_died)
			{
				/* ping failed */
				if(debug)
					printf("send_command(): server died\n");
				ret = 1;
				break;
			}
		}
		buf[rlen] = 0;

		if(rlen >= 4 && !strcmp(buf + rlen - 4, "NOK\n"))
		{
			buf[rlen-4] = 0;
			done = 1;
		}
		else if(rlen >= 3 && !strcmp(buf + rlen - 3, "OK\n"))
		{
			buf[rlen-3] = 0;
			done = 1;
		}

		if(strlen(buf) > 0)
			printf("%s", buf);
	}

	unlink(sockname);

	return(ret);
}
