/*
  capstats - daemon.c
  Copyright (C) 2009 Bert Vermeulen

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <poll.h>
#include <ifaddrs.h>
#include <sys/types.h>
#include <errno.h>
#include <signal.h>
#include <syslog.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <pcap.h>

#include "capstats.h"


/*
 * globals
 */
struct ifaddrs *system_addresses;
struct pollfd allfd[MAX_SESSIONS + 1];
nfds_t num_pfd;
struct pollevent *pollevents = NULL;
int daemon_should_die = 0;

extern int debug;
extern struct capsession *sessions;
extern int client_gid;


/*
 * command jumptable
 */
char *commands[][2] = {
		{"ping", (void *) cmd_ping},
		{"new", (void *) cmd_new},
		{"modify", (void *) cmd_modify},
		{"stop", (void *) cmd_stop},
		{"list", (void *) cmd_list},
		{"show", (void *) cmd_show},
		{"poll", (void *) cmd_poll},
		{"watch", (void *) cmd_watch},
		{NULL, NULL}
};


struct capsession *find_session_by_name(char *name)
{
	struct capsession *session, *s;

	if(name == NULL)
		return(NULL);

	session = NULL;
	for(s = sessions; s; s = s->next)
	{
		if(!strcmp(name, s->name))
		{
			session = s;
			break;
		}
	}

	return(session);
}


int cmd_ping(const char **tokens, struct sockaddr *sa, socklen_t sa_len, char **errmsg)
{

	if( (sendto(allfd[0].fd, "pong\n", 5, 0, (struct sockaddr *) sa, sa_len)) == -1)
	{
		syslog(LOG_DAEMON|LOG_ERR, "server failed to send pong response: %m");
		return(1);
	}

	return(0);
}


int cmd_new(const char **tokens, struct sockaddr *sa, socklen_t sa_len, char **errmsg)
{
	int i;
	char filter[MAX_FILTER_LEN+1];

	/* need at least a session name */
	if(tokens[1] == NULL)
	{
		*errmsg = "no session name given";
		return(1);
	}

	if(tokens[2] == NULL || strcmp(tokens[2], "filter") || tokens[3] == NULL)
	{
		*errmsg = "no filter specified";
		return(1);
	}

	/* everything after 'filter' gets concatenated into the filter */
	filter[0] = '\0';
	for(i = 3; tokens[i]; i++)
	{
		strncat(filter, tokens[i], MAX_FILTER_LEN);
		strncat(filter, " ", MAX_FILTER_LEN);
	}

	if( (init_session((char *) tokens[1], filter, errmsg)) > 0)
	{
		/* errmsg gets passed on */
		return(1);
	}

	return(0);
}


int cmd_modify(const char **tokens, struct sockaddr *sa, socklen_t sa_len, char **errmsg)
{
	struct capsession *session;
	struct bpf_program fp;
	int i;
	char filter[MAX_FILTER_LEN+1], *old_filter;

	if(tokens[1] == NULL)
	{
		*errmsg = "no session name given";
		return(1);
	}
	if(!(session = find_session_by_name( (char *) tokens[1])))
	{
		*errmsg = "session not found";
		return(1);
	}

	if(tokens[2] == NULL || strcmp(tokens[2], "filter") || tokens[3] == NULL)
	{
		*errmsg = "no filter specified";
		return(1);
	}

	/* everything after 'filter' gets concatenated into the filter */
	filter[0] = '\0';
	for(i = 3; tokens[i]; i++)
	{
		strncat(filter, tokens[i], MAX_FILTER_LEN);
		strncat(filter, " ", MAX_FILTER_LEN);
	}

	if( (pcap_compile(session->p, &fp, filter, 1, 0)) == -1)
	{
		*errmsg = pcap_geterr(session->p);
		syslog(LOG_DAEMON|LOG_ERR, "failed to compile filter: %s", *errmsg);
		return(1);
	}

	if( (pcap_setfilter(session->p, &fp)) == -1)
	{
		pcap_freecode(&fp);
		*errmsg = pcap_geterr(session->p);
		syslog(LOG_DAEMON|LOG_ERR, "failed to set filter: %s", *errmsg);
		return(1);
	}

	pcap_freecode(&fp);

	old_filter = session->filter;
	session->filter = strdup(filter);
	free(old_filter);

	return(0);
}


int cmd_stop(const char **tokens, struct sockaddr *sa, socklen_t sa_len, char **errmsg)
{
	struct capsession *session;
	int res;

	if(tokens[1] == NULL)
	{
		*errmsg = "no session name given";
		return(1);
	}

	session = find_session_by_name((char *) tokens[1]);
	if(session == NULL)
	{
		*errmsg = "session not found";
		return(1);
	}

	res = stop_session(session, errmsg);

	return(res);
}


int cmd_list(const char **tokens, struct sockaddr *sa, socklen_t sa_len, char **errmsg)
{
	struct capsession *s;
	int maxlen;
	char fmt[20], line[120];

	maxlen = 0;

	for(s = sessions; s; s = s->next)
	{
		if(strlen(s->name) > maxlen)
			maxlen = strlen(s->name);
	}

	for(s = sessions; s; s = s->next)
	{
		snprintf(fmt, 20, "%%-%ds\t%%s\n", maxlen + 2);
		snprintf(line, 120, fmt, s->name, s->filter);
		if( (sendto(allfd[0].fd, line, strlen(line), 0, (struct sockaddr *) sa, sa_len)) == -1)
		{
			syslog(LOG_DAEMON|LOG_ERR, "failed to send list response: %m");
			*errmsg = "failed to send command";
			return(1);
		}
	}

	return(0);
}


int cmd_show(const char **tokens, struct sockaddr *sa, socklen_t sa_len, char **errmsg)
{
	struct capsession *session;
	char line[256];

	if(tokens[1] == NULL)
	{
		*errmsg = "no session name given";
		return(1);
	}

	session = find_session_by_name((char *) tokens[1]);
	if(session == NULL)
	{
		*errmsg = "session not found";
		return(1);
	}

	snprintf(line, 256, "name: %s\nfilter: %s\nrunning since: %s", \
			session->name, session->filter, ctime(&(session->starttime)));
	if( (sendto(allfd[0].fd, line, strlen(line), 0, (struct sockaddr *) sa, sa_len)) == -1)
	{
		syslog(LOG_DAEMON|LOG_ERR, "failed to send show response: %m");
		*errmsg = "failed to send show response";
		return(1);
	}

	return(0);
}


int output_stats(struct capsession *session, struct sockaddr *sa, socklen_t sa_len)
{
	int ret;
	char line[256];

	snprintf(line, 255, "%s pkts-in %u pkts-out %u bytes-in %u bytes-out %u\n", \
			session->name, session->pkts_in, session->pkts_out, \
			session->bytes_in, session->bytes_out);
	if(debug)
		printf("output: %s", line);

	if( (sendto(allfd[0].fd, line, strlen(line), 0, (struct sockaddr *) sa, sa_len)) == -1)
	{
		syslog(LOG_DAEMON|LOG_ERR, "failed to send stats: %m");
		ret = 1;
	}
	else
		ret = 0;

	return(ret);
}


int cmd_poll(const char **tokens, struct sockaddr *sa, socklen_t sa_len, char **errmsg)
{
	struct capsession *session;

	if(tokens[1] == NULL)
	{
		*errmsg = "no session name given";
		return(1);
	}

	session = find_session_by_name((char *) tokens[1]);
	if(session == NULL)
	{
		*errmsg = "session not found";
		return(1);
	}

	output_stats(session, sa, sa_len);

	return(0);
}


int cmd_watch(const char **tokens, struct sockaddr *sa, socklen_t sa_len, char **errmsg)
{
	struct capsession *session;
	struct pollevent *pe, *new_pe;
	int frequency;

	/* find session */
	if(tokens[1] == NULL)
	{
		*errmsg = "no session name specified";
		return(1);
	}
	session = find_session_by_name((char *) tokens[1]);
	if(session == NULL)
	{
		*errmsg = "session not found";
		return(1);
	}

	/* watch with which frequency? */
	if(tokens[2] == NULL || tokens[3] == NULL || strcmp(tokens[2], "frequency"))
	{
		*errmsg = "no frequency specified";
		return(1);
	}
	frequency = atoi(tokens[3]);
	if(frequency == 0)
	{
		*errmsg = "invalid frequency";
		return(1);
	}

	/* set up poll event */
	new_pe = malloc(sizeof(struct pollevent));
	memset(new_pe, 0, sizeof(struct pollevent));
	new_pe->session = session;
	new_pe->polltime = frequency;
	new_pe->sa = malloc(sa_len);
	new_pe->frequency = frequency;
	memcpy(new_pe->sa, sa, sa_len);
	new_pe->sa_len = sa_len;

	if(pollevents == NULL)
	{
		pollevents = new_pe;
	}
	else
	{
		for(pe = pollevents; pe->next; pe = pe->next);
		pe->next = new_pe;
	}

	output_stats(session, sa, sa_len);
	alarm(frequency);

	return(-1);
}


char *show_addr(int family, char *addr)
{
	static char addrbuf[INET6_ADDRSTRLEN+1];
	char *famaddr;

	memset(addrbuf, 0, sizeof(addrbuf));
	if(family == AF_INET || family == AF_INET6)
	{
		if(family == AF_INET)
			famaddr = (char *) &((struct sockaddr_in *)addr)->sin_addr;
		else if(family == AF_INET6)
			famaddr = (char *) &((struct sockaddr_in6 *)addr)->sin6_addr;
		inet_ntop(family, famaddr, addrbuf, sizeof(addrbuf));
	}

	return(addrbuf);
}


void list_addrs()
{
	struct ifaddrs *ifap;
	int family;

	ifap = system_addresses;
	while(ifap->ifa_next)
	{
		family = ifap->ifa_addr->sa_family;
		if(family == AF_INET || family == AF_INET6)
		{
			printf("interface %s ", ifap->ifa_name);
			printf("family %d ", ifap->ifa_addr->sa_family);
			printf("addr %s\n", show_addr(family, (char *)ifap->ifa_addr));
		}
		ifap = ifap->ifa_next;
	}

}


int setup_cmdsocket()
{
	struct sockaddr_un listensa;
	struct pollfd *pfd;
	int cmdsock, true;

	unlink(CMDSOCKET);

	if( (cmdsock = socket(AF_UNIX, SOCK_DGRAM, 0)) == -1)
	{
		syslog(LOG_DAEMON|LOG_ERR, "failed to create socket: %m");
		printf("failed to create socket: %s\n", strerror(errno));
		return(1);
	}
	if(debug)
		printf("server: created command socket\n");

	true = 1;
	if(setsockopt(cmdsock, SOL_SOCKET, SO_PASSCRED, &true, sizeof(true)) == -1)
	{
		syslog(LOG_DAEMON|LOG_ERR, "failed to set SO_PASSCRED on socket: %m");
		printf("failed to set SO_PASSCRED on socket: %s\n", strerror(errno));
		close(cmdsock);
		return(1);
	}

	memset(&listensa, 0, sizeof(listensa));
	listensa.sun_family = AF_UNIX;
	strcpy(listensa.sun_path, CMDSOCKET);
	if(bind(cmdsock, (const struct sockaddr *) &listensa, sizeof(listensa)) == -1)
	{
		syslog(LOG_DAEMON|LOG_ERR, "failed to bind command socket: %m");
		printf("failed to bind command socket: %s\n", strerror(errno));
		close(cmdsock);
		return(1);
	}
	if(debug)
		printf("server: command socket bound\n");

	chmod(CMDSOCKET, 0666);

	/* first FD polled is always the command socket */
	pfd = &allfd[0];
	pfd->fd = cmdsock;
	pfd->events = POLLIN;
	pfd->revents = 0;
	num_pfd = 1;

	return(0);
}


void handle_command(char *cmd, struct sockaddr *sa, int sa_len)
{
	int (*func)(char **tokens, struct sockaddr *sa, socklen_t sa_len, char **errmsg);
	int ret, i;
	char response[1024], **tokens, *errmsg;

	if(debug)
		printf("server command socket received [%s]\n", cmd);

	tokens = tokenize(cmd);
	if(tokens == NULL)
	{
		if( (sendto(allfd[0].fd, "NOK\n", 4, 0, sa, sa_len)) == -1)
		{
			syslog(LOG_DAEMON|LOG_ERR, "server failed to send response: %m (NOK)");
		}
		return;
	}

	snprintf(response, 1023, "unknown command %s\nNOK\n", tokens[0]);
	for(i = 0; commands[i][0] != NULL; i++)
	{
		if(!strcmp(tokens[0], commands[i][0]))
		{
			func = (void *) commands[i][1];
			ret = (*func)(tokens, sa, sa_len, &errmsg);
			if(ret > 0)
			{
				strcpy(response, errmsg);
				if(response[strlen(response)] != '\n')
					strcat(response, "\n");
				strcat(response, "NOK\n");
			}
			else if(ret == 0)
				strcpy(response, "OK\n");
			else
			{
				/* client should keep running */
				response[0] = '\0';
			}
			break;
		}
	}

	if( (sendto(allfd[0].fd, response, strlen(response), 0, sa, sa_len)) == -1)
	{
		syslog(LOG_DAEMON|LOG_ERR, "server failed to send response: %m (%s)", response);
	}

	/* the buffer and the pointers to the tokens inside it */
	free(tokens[0]);
	free(tokens);

}


void pollrun(int signum)
{
	struct pollevent *pe, *this_pe, *p;
	time_t now;
	unsigned int next_run;
	int res;

	res = 0;
	now = time(NULL);
	next_run = 3600;
	this_pe = NULL;
	for(pe = pollevents; pe; pe = pe->next)
	{
		if(now >= pe->polltime)
		{
			/* save in case we need to delete this later */
			this_pe = pe;
			/* this pollevent is (over)due for a stats line */
			res = output_stats(pe->session, pe->sa, pe->sa_len);
			/* set up next poll run */
			pe->polltime = now + pe->frequency;
		}
		if(res == 0 && pe->polltime - now < next_run)
			next_run = pe->polltime - now;
	}

	if(res > 0)
	{
		/* something went wrong sending stats to the watching socket.
		 * assume the client went away, and delete this pollevent.
		 */
		if(debug)
			printf("client went away, deleting pollevent\n");

		if(this_pe == pollevents)
		{
			pollevents = this_pe->next;
		}
		else
		{
			for(p = pollevents; p; p = p->next)
			{
				if(p->next == this_pe)
				{
					p->next = this_pe->next;
					break;
				}
			}
		}
		free(this_pe->sa);
		free(this_pe);
	}
	else
	{
		/* all went well, set up the next poll run */
		alarm(next_run);
	}

}


int get_creds(int sock, struct ucred *cred)
{
	struct msghdr msg;
	struct cmsghdr *cmsg;
	struct iovec iov;
	int ret;
	char control[512];
	char buf[512];

	iov.iov_base = &buf;
	iov.iov_len = sizeof(buf);
	memset(&msg, 0, sizeof(msg));
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = &control;
	msg.msg_controllen = sizeof(control);

	/* peek ahead at the incoming data and extract its credentials */
	if( (recvmsg(sock, &msg, MSG_PEEK)) == -1 || msg.msg_flags != 0)
	{
		syslog(LOG_DAEMON|LOG_ERR, "failed to get credentials: %m (flags 0x%.2x)", msg.msg_flags);
		printf("failed to get credentials: %s (flags 0x%.2x)\n", strerror(errno), msg.msg_flags);
		return(0);
	}

	ret = 0;
	for(cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg))
	{
		if(cmsg->cmsg_level == SOL_SOCKET && cmsg->cmsg_type == SCM_CREDENTIALS) {
			memcpy(cred, CMSG_DATA(cmsg), sizeof(struct ucred));
			ret = 1;
			break;
		}
	}
	if(debug)
		printf("got credentials: pid %d uid %d gid %d\n", cred->pid, cred->uid, cred->gid);

	return(ret);
}


int check_creds(int sock)
{
	struct ucred cred;
	struct passwd *pw;
	struct group *gr;
	int i;

	if(!get_creds(allfd[0].fd, &cred))
		return(0);

	/* the right primary group, or root */
	if(client_gid == cred.gid || cred.uid == 0)
		return(1);

	/* check secondary groups */
	if( (pw = getpwuid(cred.uid)) )
		if( (gr = getgrgid(client_gid)) )
			for(i = 0; gr->gr_mem[i]; i++)
				if(!strcmp(pw->pw_name, gr->gr_mem[i]))
					/* connecting user is a member of this group */
					return(1);

	syslog(LOG_DAEMON|LOG_ERR, "command socket permission denied for uid %d", cred.uid);

	return(0);
}


int start_daemon()
{
	struct sockaddr_un cmdsa;
	struct capsession *s;
	socklen_t cmdsa_len;
	int ret, res, rlen, permission, i;
	u_char buf[1522];

	pollevents = NULL;

	if(getifaddrs(&system_addresses))
	{
		syslog(LOG_DAEMON|LOG_ERR, "unable to find local interface addresses: %m");
		printf("unable to find local interface addresses: %s\n", strerror(errno));
		return(1);
	}
	if(debug)
		list_addrs();

	/* set up syslogging */
	openlog(PROGNAME, 0, LOG_DAEMON);

	if(setup_cmdsocket())
		return(1);

	pollevents = NULL;
	signal(SIGALRM, pollrun);

	signal(SIGINT, kill_daemon);
	signal(SIGTERM, kill_daemon);

	load_sessions();

	while(1)
	{
		res = poll(allfd, num_pfd, -1);
		if(res == -1)
		{
			if(errno == EINTR)
			{
				if(daemon_should_die)
					/* daemon got SIGINT or SIGTERM, exit cleanly */
					break;
				else
					/* SIGALRM interrupted for an output run */
					continue;
			}
			else
			{
				syslog(LOG_DAEMON|LOG_ERR, "poll failed: %m");
				ret = 1;
				break;
			}
		}
		if(allfd[0].revents & POLLIN)
		{
			/* something came in on the command socket */
			permission = check_creds(allfd[0].fd);
			cmdsa_len = sizeof(struct sockaddr_un);
			rlen = recvfrom(allfd[0].fd, buf, sizeof(buf)-1, 0, (struct sockaddr *) &cmdsa, &cmdsa_len);
			if(!permission)
				sendto(allfd[0].fd, MSG_PERMDENIED, strlen(MSG_PERMDENIED), 0, (struct sockaddr *) &cmdsa, cmdsa_len);
			else
			{
				if(rlen > -1)
				{
					buf[rlen] = 0;
					handle_command((char *) buf, (struct sockaddr *) &cmdsa, cmdsa_len);
				}
				else
					syslog(LOG_DAEMON|LOG_ERR, "server failed to receive command: %m");
			}
		}

		for(i = 1; i <= num_pfd; i++)
		{
			if(allfd[i].revents & POLLIN)
			{
				rlen = recvfrom(allfd[i].fd, buf, sizeof(buf)-1, 0, NULL, NULL);
				if(rlen == -1)
					continue;

				/* find out which session this FD belongs to */
				for(s = sessions; s; s = s->next)
				{
					if(s->fd == allfd[i].fd)
					{
						proc_packet(s, buf, rlen);
						break;
					}
				}
			}
		}
	}

	save_sessions();
	save_session_stats();
	stop_all_sessions();

	freeifaddrs(system_addresses);
	unlink(CMDSOCKET);

	return(0);
}


void kill_daemon(int signum)
{

	/* make the main poll loop exit cleanly */
	daemon_should_die = 1;

}


