/*
  capstats - session.c
  Copyright (C) 2009 Bert Vermeulen

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ifaddrs.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <poll.h>
#include <time.h>
#include <syslog.h>
#include <errno.h>

#include <pcap.h>

#include "capstats.h"

struct capsession *sessions = NULL;

extern struct ifaddrs *system_addresses;
extern struct pollfd allfd[];
extern int num_pfd;
extern int debug;
extern struct pollevent *pollevents;


int get_payload_offset(linktype)
{
	int offset;

	switch(linktype)
	{
	case DLT_EN10MB:
		offset = 14;
		break;
	default:
		offset = 0;
	}

	return(offset);
}


/*
 * returns 0 if not found, 1 if srcaddr is one of ours, 2 is dstaddr is
 */
int identify_address(int family, char *srcaddr, char *dstaddr)
{
	struct ifaddrs *ifap;
	int result;
	char *myaddr;

	result = 0;
	ifap = system_addresses;
	while(ifap->ifa_next)
	{
		if(ifap->ifa_addr->sa_family == family)
		{
			if(family == AF_INET)
			{
				myaddr = (char *) &((struct sockaddr_in *)ifap->ifa_addr)->sin_addr;
				if(memcmp(srcaddr, myaddr, 4) == 0)
				{
					result = 1;
					break;
				}
				else if(memcmp(dstaddr, myaddr, 4) == 0)
				{
					result = 2;
					break;
				}
			}
			else if(family == AF_INET6)
			{
				if(memcmp(srcaddr, &((struct sockaddr_in6 *)ifap->ifa_addr)->sin6_addr, 16) == 0)
				{
					result = 1;
					break;
				}
			}
		}
		ifap = ifap->ifa_next;
	}

	return(result);
}


void proc_packet(struct capsession *session, const u_char *pkt, int pkt_len)
{
	uint16_t rawtype, ethertype;
	int ident;
	char *addr;

	session->pkts++;
	if(session->linktype == DLT_EN10MB)
	{
		rawtype = *(pkt + OFFSET_ETH_ETHERTYPE);
		ethertype = ntohs(rawtype);
		if(ethertype == 0x0800)
		{
			addr = (char *) pkt + OFFSET_ETH_PAYLOAD + OFFSET_IP_SRCADDR;
			ident = identify_address(AF_INET, addr, addr + 4);
			if(ident == 1)
			{
				session->pkts_out++;
				session->bytes_out += pkt_len - OFFSET_ETH_PAYLOAD;
			}
			else if(ident == 2)
			{
				session->pkts_in++;
				session->bytes_in += pkt_len - OFFSET_ETH_PAYLOAD;
			}
		}
	}

	if(debug)
		printf("captured packet: pkts %u in %u out %u bytes in %u bytes out %u\n", session->pkts, \
				session->pkts_in, session->pkts_out, \
				session->bytes_in, session->bytes_out);
}


int init_session(char *session_name, char *filter, char **errmsg)
{
	struct bpf_program fp;
	struct capsession *session, *s;
	pcap_t *p;
	char errbuf[PCAP_ERRBUF_SIZE];
	char *ifname;

	if(num_pfd > MAX_SESSIONS)
	{
		*errmsg = "maximum number of sessions reached";
		return(1);
	}

	for(s = sessions; s; s = s->next)
	{
		if(!strcmp(session_name, s->name))
		{
			*errmsg = "a session by that name already exists";
			return(1);
		}
	}

	if(debug)
		printf("init session [%s] filter [%s]\n", session_name, filter);

	ifname = pcap_lookupdev(errbuf);
	if(!ifname)
	{
		syslog(LOG_DAEMON|LOG_ERR, "failed to find a suitable device: %s", errbuf);
		*errmsg = "failed to find a suitable device";
		return(1);
	}

	errbuf[0] = '\0';
	p = pcap_open_live(ifname, SNAPLEN, 0, CAPDELAY, errbuf);
	if(!p)
	{
		syslog(LOG_DAEMON|LOG_ERR, "failed to open pcap session: %s", errbuf);
		*errmsg = "failed to open pcap session";
		return(1);
	}

	if( (pcap_compile(p, &fp, filter, 1, 0)) == -1)
	{
		*errmsg = pcap_geterr(p);
		syslog(LOG_DAEMON|LOG_ERR, "failed to compile filter: %s", *errmsg);
		return(1);
	}

	if( (pcap_setfilter(p, &fp)) == -1)
	{
		*errmsg = pcap_geterr(p);
		syslog(LOG_DAEMON|LOG_ERR, "failed to set filter: %s", *errmsg);
		return(1);
	}

	/* done with the compiled filter */
	pcap_freecode(&fp);

	session = malloc(sizeof(struct capsession));
	memset(session, 0, sizeof(struct capsession));
	session->p = p;
	snprintf(session->name, MAX_SESSION_NAME_LENGTH-1, "%s", session_name);
	session->filter = strdup(filter);
	session->linktype = pcap_datalink(p);
	session->payload_offset = get_payload_offset(session->linktype);
	session->fd = pcap_get_selectable_fd(p);
	session->starttime = time(NULL);
	if(sessions)
	{
		for(s = sessions; s->next; s = s->next);
		s->next = session;
	}
	else
		sessions = session;

	load_session_stats(session);

	allfd[num_pfd].fd = session->fd;
	allfd[num_pfd].events = POLLIN;
	allfd[num_pfd].revents = 0;
	num_pfd++;

	return(0);
}


int stop_session(struct capsession *session, char **errmsg)
{
	struct capsession *s;
	struct pollevent *pe, *prev_pe;
	int old, new;

	/* get rid of any poll events using this session */
	pe = NULL;
	if(pollevents && pollevents->session == s)
	{
		pe = pollevents;
		pollevents = pollevents->next;
	}
	else
	{
		for(pe = pollevents, prev_pe = pollevents; pe; pe = pe->next)
		{
			if(pe->session == s)
			{
				prev_pe->next = pe->next;
				break;
			}
			prev_pe = pe;
		}
	}
	if(pe && pe->session == s)
	{
		/* tell the client listening to this watch to disconnect */
		sendto(allfd[0].fd, "OK\n", 3, 0, pe->sa, pe->sa_len);
		free(pe);
	}

	/* rebuild the pollfd list without this session's fd in it */
	for(old = 0, new = 0; old < num_pfd; old++)
		if(allfd[old].fd != session->fd)
			allfd[new++] = allfd[old];
	num_pfd--;

	pcap_close(session->p);

	/* free session resources */
	if(sessions == session)
		sessions = session->next;
	else
	{
		for(s = sessions; s->next == session; s = s->next);
		s->next = session->next;
	}
	free(session->filter);
	free(session);

	return(0);
}


void stop_all_sessions(void)
{
	struct capsession *s, *s_next;
	char *errmsg;

	for(s = sessions; s; s = s_next)
	{
		s_next = s->next;
		stop_session(s, &errmsg);
	}

}


void load_sessions(void)
{
	FILE *sessionsfile;
	int res, cnt, i;
	char buf[MAX_FILTER_LEN * 2], filter[MAX_FILTER_LEN+1], *errmsg, **tokens;

	if( !(sessionsfile = fopen(SESSIONSFILE, "r")) )
		return;

	cnt = 1;
	while(fgets(buf, sizeof(buf), sessionsfile))
	{
		tokens = tokenize(buf);
		if(tokens[0] && tokens[1] && !strcmp(tokens[1], "filter") && tokens[2])
		{
			/* looks good... <sessionname> filter <filter> */
			filter[0] = '\0';
			for(i = 2; tokens[i]; i++)
			{
				strncat(filter, tokens[i], MAX_FILTER_LEN);
				strncat(filter, " ", MAX_FILTER_LEN);
			}
			res = init_session(tokens[0], filter, &errmsg);
			if(res)
			{
				syslog(LOG_DAEMON|LOG_ERR, "unable to initialize session %s: %s", tokens[0], errmsg);
			}
		}
		else
		{
			syslog(LOG_DAEMON|LOG_ERR, "invalid saved session in line %d of %s", cnt, SESSIONSFILE);
		}
		cnt++;
	}

	fclose(sessionsfile);

}


void save_sessions(void)
{
	FILE *sessionsfile;
	struct stat st;
	struct capsession *s;

	if(stat(PERMDIR, &st))
	{
		mkdir(PERMDIR, 0755);
	}

	if( !(sessionsfile = fopen(SESSIONSFILE, "w")) )
	{
		syslog(LOG_DAEMON|LOG_ERR, "unable to open %s for writing: %s", SESSIONSFILE, strerror(errno));
		return;
	}

	for(s = sessions; s; s = s->next)
	{
		fprintf(sessionsfile, "%s filter %s\n", s->name, s->filter);
	}
	fclose(sessionsfile);

}


void load_session_stats(struct capsession *session)
{
	FILE *statsfile;
	int res;
	char fmt[256], line[128];

	if( (statsfile = fopen(STATSFILE, "r")) )
	{
		snprintf(fmt, 255, "%s pkts-in %%u pkts-out %%u bytes-in %%u bytes-out %%u\n", session->name);
		while(fgets(line, 127, statsfile))
		{
			res = sscanf(line, fmt, &session->pkts_in, &session->pkts_out, \
						&session->bytes_in, &session->bytes_out);
			if(res > 0)
				break;
		}
	}

}


void save_session_stats(void)
{
	struct stat st;
	FILE *statsfile;
	struct capsession *session;

	if(stat(PERMDIR, &st))
	{
		mkdir(PERMDIR, 0755);
	}

	if( !(statsfile = fopen(STATSFILE, "w")) )
	{
		syslog(LOG_DAEMON|LOG_ERR, "unable to open %s for writing: %s", STATSFILE, strerror(errno));
		return;
	}

	for(session = sessions; session; session = session->next)
	{
		fprintf(statsfile, "%s pkts-in %d pkts-out %d bytes-in %d bytes-out %d\n", \
				session->name, session->pkts_in, session->pkts_out, \
				session->bytes_in, session->bytes_out);
	}
	fclose(statsfile);

}


